/***************************************************************************
 *   Copyright (C) 2004-2024 by Pere Constans
 *   constans@molspaces.com
 *   cb2Bib version 2.0.2. Licensed under the GNU GPL version 3.
 *   See the LICENSE file that comes with this distribution.
 ***************************************************************************/
#ifndef PORTING_TO_QT_6_H
#define PORTING_TO_QT_6_H

#include <QtGlobal>
#if QT_VERSION >= QT_VERSION_CHECK(5, 14, 0)
#define _QT_6_KeepEmptyParts Qt::KeepEmptyParts
#define _QT_6_SkipEmptyParts Qt::SkipEmptyParts
#define _QT_6_endl Qt::endl
#else
#define _QT_6_KeepEmptyParts QString::KeepEmptyParts
#define _QT_6_SkipEmptyParts QString::SkipEmptyParts
#define _QT_6_endl endl
#endif

#endif
